from PyQt5.QtWidgets import QMenu
from qtpy.QtWidgets import (
    QHBoxLayout,
    QWidget,
    QPushButton,
    QToolButton,
    QStyle,
    QGroupBox,
    QLabel,
)
from qtpy.QtCore import Qt

from aydin.io.datasets import examples_single


class QProgramFlowDiagramWidget(QWidget):
    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.parent = parent
        self.highlightable_buttons = []

        self.layout = QHBoxLayout()

        # self.main_group_box = QGroupBox("Program Flow")
        self.layout = QHBoxLayout()
        self.layout.setSpacing(5)

        self.load_data_group_box = QGroupBox("Load data")
        self.load_data_group_box_layout = QHBoxLayout()
        self.load_data_group_box_layout.setSpacing(0)
        self.load_data_group_box_layout.setContentsMargins(3, -1, 3, -1)

        self.add_files_button = QPushButton("Add File(s)")
        self.load_data_group_box_layout.addWidget(self.add_files_button)

        self.load_data_group_box_layout.addWidget(QLabel("or"))

        self.load_sample_image_button = QPushButton("Load example")
        menu = QMenu()
        menu_items = {
            "Fountain": examples_single.fountain,
            "Mona Lisa": examples_single.monalisa,
            "Gauss": examples_single.gauss_noisy,
            "Periodic": examples_single.periodic_noise,
            "Chessboard": examples_single.brown_chessboard_gray,
            "HCR (Royer)": examples_single.royerlab_hcr,
            "Blastocyst Fracking (Maitre)": examples_single.maitre_mouse,
        }
        for item in menu_items.keys():
            action = menu.addAction(item)
            action.setIconVisibleInMenu(False)
        menu.triggered.connect(
            lambda x: self.parent.load_sample_image(menu_items[x.text()])
        )

        self.load_sample_image_button.setMenu(menu)
        self.load_data_group_box_layout.addWidget(self.load_sample_image_button)

        self.load_data_group_box.setLayout(self.load_data_group_box_layout)
        self.layout.addWidget(self.load_data_group_box)

        self.layout.addWidget(self.forward_button())

        # choose image options
        self.choose_image_options_group_box = QGroupBox("Choose image options")
        self.choose_image_options_group_box_layout = QHBoxLayout()
        self.choose_image_options_group_box_layout.setSpacing(5)
        self.choose_image_options_group_box_layout.setContentsMargins(3, -1, 3, -1)

        self.files_button = QPushButton("File(s)")
        self.highlightable_buttons.append(self.files_button)
        self.choose_image_options_group_box_layout.addWidget(self.files_button)

        self.images_button = QPushButton("Image(s)")
        self.highlightable_buttons.append(self.images_button)
        self.choose_image_options_group_box_layout.addWidget(self.images_button)

        self.dimensions_button = QPushButton("Dimensions")
        self.highlightable_buttons.append(self.dimensions_button)
        self.choose_image_options_group_box_layout.addWidget(self.dimensions_button)

        self.training_crop_button = QPushButton("Training Crop")
        self.highlightable_buttons.append(self.training_crop_button)
        self.choose_image_options_group_box_layout.addWidget(self.training_crop_button)

        self.inference_crop_button = QPushButton("Denoising Crop")
        self.highlightable_buttons.append(self.inference_crop_button)
        self.choose_image_options_group_box_layout.addWidget(self.inference_crop_button)

        self.choose_image_options_group_box.setLayout(
            self.choose_image_options_group_box_layout
        )

        self.layout.addWidget(self.choose_image_options_group_box)

        self.layout.addWidget(self.forward_button())

        self.processing_group_box = QGroupBox("Process")
        self.processing_group_box_layout = QHBoxLayout()
        self.processing_group_box_layout.setSpacing(5)
        self.processing_group_box_layout.setContentsMargins(3, -1, 3, -1)

        self.preprocess_button = QPushButton("Preprocessing")
        self.highlightable_buttons.append(self.preprocess_button)
        self.processing_group_box_layout.addWidget(self.preprocess_button)

        self.processing_group_box_layout.addWidget(self.forward_button())

        self.denoise_button = QPushButton("Denoise")
        self.highlightable_buttons.append(self.denoise_button)
        self.processing_group_box_layout.addWidget(self.denoise_button)

        self.processing_group_box_layout.addWidget(self.forward_button())

        self.postprocess_button = QPushButton("Postprocessing")
        self.highlightable_buttons.append(self.postprocess_button)
        self.processing_group_box_layout.addWidget(self.postprocess_button)

        self.processing_group_box.setLayout(self.processing_group_box_layout)

        self.layout.addWidget(self.processing_group_box)

        self.layout.setAlignment(Qt.AlignHCenter)

        self.setLayout(self.layout)

    def highlight_button(self, current_tab_name):
        self.reset_buttons()
        for button in self.highlightable_buttons:
            if current_tab_name.lower() in button.text().lower():
                button.setStyleSheet(
                    "background-color: qlineargradient( x1:0 y1:0, x2:1 y2:0, stop:0 #0897c7, stop:1 #586727)"
                )
        else:
            if "processing" in current_tab_name.lower():
                self.preprocess_button.setStyleSheet(
                    "background-color: qlineargradient( x1:0 y1:0, x2:1 y2:0, stop:0 #0897c7, stop:1 #586727)"
                )
                self.postprocess_button.setStyleSheet(
                    "background-color: qlineargradient( x1:0 y1:0, x2:1 y2:0, stop:0 #0897c7, stop:1 #586727)"
                )

    def reset_buttons(self):
        for button in self.highlightable_buttons:
            button.setStyleSheet("")

    def forward_button(self):
        button = QToolButton()
        button.setEnabled(False)
        button.setIcon(button.style().standardIcon(QStyle.SP_MediaSeekForward))

        return button



import time

import numpy
import pytest
import tensorflow as tf  # noqa: F401
from skimage.data import camera
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

# from tensorflow_core.python.keras.backend import clear_session
from tensorflow.python.keras.backend import clear_session

from aydin.io import io
from aydin.io.datasets import normalise, add_noise, examples_single
from aydin.it.cnn import ImageTranslatorCNN


def test_it_cnn_history():
    """
    Check if training history is properly recorded.
    """
    start = time.time()
    max_epochs = 2
    data = numpy.zeros((64, 64))
    it = ImageTranslatorCNN(
        model_architecture="unet",
        training_architecture='checkran',
        nb_unet_levels=1,
        patch_size=64,
        batch_size=1,
        mask_size=3,
        total_num_patches=1,
        patience=1,
        max_epochs=max_epochs,
    )
    it.train(data, data)
    history = it.loss_history
    for key, val in history.history.items():
        assert len(val) == max_epochs
    assert len(history.epoch) == max_epochs

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    clear_session()


def test_it_cnn_shiftconv_light():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    start = time.time()
    max_epochs = 30
    image_width = 100
    image = normalise(camera())
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    noisy = add_noise(image)

    print("noisy shape: ", noisy.shape)

    it = ImageTranslatorCNN(
        model_architecture="unet",
        training_architecture='shiftconv',
        nb_unet_levels=2,
        batch_norm=None,  # 'instance',
        max_epochs=max_epochs,
    )
    it.train(noisy, noisy)
    denoised = it.translate(noisy, tile_size=image_width)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy

    clear_session()


def test_it_cnn_checkerbox_light():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    start = time.time()
    max_epochs = 5
    image_width = 100
    image = normalise(camera())
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        model_architecture="unet",
        training_architecture='checkerbox',
        nb_unet_levels=2,
        mask_size=3,
        batch_norm='instance',
        max_epochs=max_epochs,
    )
    it.train(noisy, noisy)
    denoised = it.translate(noisy, tile_size=image_width)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy
    clear_session()


def test_it_cnn_random_light():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    start = time.time()
    max_epochs = 5
    image_width = 100
    image = normalise(camera())
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        model_architecture="unet",
        training_architecture='random',
        nb_unet_levels=2,
        batch_norm='instance',
        max_epochs=max_epochs,
    )
    it.train(noisy, noisy)
    denoised = it.translate(noisy, tile_size=image_width)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > psnr_noisy * 0.9 and ssim_denoised > ssim_noisy * 0.9
    clear_session()


def test_it_cnn_checkran_light():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    start = time.time()
    max_epochs = 5
    image_width = 100

    image = normalise(camera())
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    # Test with arbitrary input shape
    arbitrary_shape = (1, 1) + image.shape
    batch_dims = tuple([True if i == 1 else False for i in arbitrary_shape])
    image = image.reshape(arbitrary_shape)
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        model_architecture="unet",
        training_architecture='checkran',
        nb_unet_levels=2,
        mask_size=3,
        batch_norm='instance',
        max_epochs=max_epochs,
    )
    it.train(noisy, noisy, batch_axes=batch_dims)
    denoised = it.translate(noisy, tile_size=image_width, batch_axes=batch_dims)
    assert denoised.shape == noisy.shape
    denoised = denoised.squeeze()
    noisy = noisy.squeeze()
    image = image.squeeze()

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy
    clear_session()


def test_it_cnn_jinet2D_light():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    start = time.time()
    max_epochs = 30
    image_width = 100
    image = normalise(camera())
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        model_architecture='jinet', patch_size=image_width, max_epochs=max_epochs
    )
    it.train(noisy, noisy)
    denoised = it.translate(noisy, tile_size=image_width)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy
    clear_session()


def test_it_cnn_jinet2D_supervised_light():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    start = time.time()
    max_epochs = 30
    image_width = 100
    image = normalise(camera())
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        model_architecture='jinet', patch_size=image_width, max_epochs=max_epochs
    )
    it.train(noisy, image)
    denoised = it.translate(noisy, tile_size=image_width)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy
    clear_session()


def test_it_cnn_jinet3D_light():
    """
    Demo for self-supervised denoising
    """
    start = time.time()
    max_epochs = 30
    image_width = 64
    image_path = examples_single.royerlab_hcr.get_path()
    image, metadata = io.imread(image_path)
    image = image[10:20, 1:2, 100 : 100 + image_width, 200 : 200 + image_width]
    image = rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        model_architecture='jinet', patch_size=image_width, max_epochs=max_epochs
    )
    it.train(
        noisy, noisy, batch_axes=metadata.batch_axes, channel_axes=metadata.channel_axes
    )
    denoised = it.translate(
        noisy,
        tile_size=image_width,
        batch_axes=metadata.batch_axes,
        channel_axes=metadata.channel_axes,
    )

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    noisy = numpy.squeeze(noisy)
    image = numpy.squeeze(image)
    denoised = numpy.squeeze(denoised)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > (psnr_noisy * 0.5) and ssim_denoised > (ssim_noisy * 0.5)
    clear_session()


def test_it_cnn_jinet3D_supervised_light():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    start = time.time()
    max_epochs = 30
    image_width = 64
    image_path = examples_single.royerlab_hcr.get_path()
    image, metadata = io.imread(image_path)
    image = image[10:20, 1:2, 100 : 100 + image_width, 200 : 200 + image_width]
    image = rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        model_architecture='jinet', patch_size=image_width, max_epochs=max_epochs
    )
    it.train(
        noisy, image, batch_axes=metadata.batch_axes, channel_axes=metadata.channel_axes
    )
    denoised = it.translate(
        noisy,
        tile_size=image_width,
        batch_axes=metadata.batch_axes,
        channel_axes=metadata.channel_axes,
    )

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    noisy = numpy.squeeze(noisy)
    image = numpy.squeeze(image)
    denoised = numpy.squeeze(denoised)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image, multichannel=True)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image, multichannel=True)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > (psnr_noisy * 0.5) and ssim_denoised > (ssim_noisy * 0.5)
    clear_session()


@pytest.mark.heavy
def test_it_cnn_random_patching():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    start = time.time()
    max_epochs = 16
    image_width = 100
    image = normalise(camera())
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        training_architecture='random',
        nb_unet_levels=2,
        batch_norm='instance',
        max_epochs=max_epochs,
        patch_size=64,
    )
    it.train(noisy, noisy)
    denoised = it.translate(noisy, tile_size=image_width)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    stop = time.time()
    print(f"Total elapsed time: {stop - start} ")
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy



import numpy
from scipy.ndimage import shift
from skimage.exposure import rescale_intensity

from aydin.util.fast_shift.fast_shift import fast_shift
from aydin.io.datasets import newyork, examples_single


def _normalise(image):
    return rescale_intensity(
        image.astype(numpy.float32, copy=False), in_range='image', out_range=(0, 1)
    )


def test_fast_shift_filter_type_support():
    image = newyork()

    _run_test_for_type(image.astype(dtype=numpy.float32))
    _run_test_for_type(image.astype(dtype=numpy.float16), decimal=1)
    _run_test_for_type(image.astype(dtype=numpy.uint32), decimal=0)
    _run_test_for_type(image.astype(dtype=numpy.uint16), decimal=0)
    _run_test_for_type(image.astype(dtype=numpy.uint8), decimal=0)


def _run_test_for_type(image, decimal=3, _shift=(-1, 3)):
    shifted_image = fast_shift(image, shift=_shift)
    assert shifted_image.dtype == image.dtype

    shifted_image = shifted_image.astype(dtype=numpy.float32, copy=False)
    scipy_shifted_image = shift(
        image.astype(dtype=numpy.float32, copy=False), shift=_shift, mode="constant"
    ).astype(dtype=numpy.float32, copy=False)
    numpy.testing.assert_array_almost_equal(
        shifted_image, scipy_shifted_image, decimal=decimal
    )


def test_compute_uniform_filter_different_sizes():

    shifts = [(1, 2), (16, 7), (3, 7), (4, 4)]

    for current_shift in shifts:
        test_fast_shift_filter_2d(_shift=current_shift)


def test_fast_shift_filter_1d(_shift=(-1,)):
    image = _normalise(newyork().astype(numpy.float32))
    image = image[512, :]

    shifted_image = fast_shift(image, shift=_shift)
    scipy_shifted_image = shift(image, shift=_shift, mode="constant")

    numpy.testing.assert_array_almost_equal(
        shifted_image, scipy_shifted_image, decimal=3
    )


def test_fast_shift_filter_2d(_shift=(-1, 3)):
    image = _normalise(newyork().astype(numpy.float32))

    shifted_image = fast_shift(image, shift=_shift)
    scipy_shifted_image = shift(image, shift=_shift, mode="constant")

    numpy.testing.assert_array_almost_equal(
        shifted_image, scipy_shifted_image, decimal=3
    )


def test_fast_shift_filter_3d(_shift=(-1, 3, -7)):
    islet = examples_single.royerlab_hcr.get_array().squeeze()
    image = islet[2, :60, 0:256, 0:256]

    shifted_image = fast_shift(image, shift=_shift)
    scipy_shifted_image = shift(image, shift=_shift, mode="constant")

    numpy.testing.assert_array_almost_equal(
        shifted_image, scipy_shifted_image, decimal=3
    )


def test_fast_shift_filter_4d(_shift=(-1, 3, -7, +13)):
    image = examples_single.hyman_hela.get_array().squeeze()
    image = image[0:10, 0:10, 0:128, 0:128]

    shifted_image = fast_shift(image, shift=_shift)
    scipy_shifted_image = shift(image, shift=_shift, mode="constant")

    numpy.testing.assert_array_almost_equal(
        shifted_image, scipy_shifted_image, decimal=3
    )



import torch
from torch import nn


class PSFConvolutionLayer(nn.Module):
    def __init__(self, kernel_psf, num_channels_in=1, num_channels_out=1):
        kernel_size = kernel_psf.shape[0]
        super().__init__()
        self.seq = nn.Sequential(
            nn.ReflectionPad2d((kernel_size - 1) // 2),
            nn.Conv2d(
                num_channels_in,
                num_channels_out,
                kernel_size,
                stride=1,
                padding=0,
                bias=False,
                groups=num_channels_in,
            ),
        )

        self.weights_init(kernel_psf)

    def weights_init(self, kernel_psf):
        for name, f in self.named_parameters():
            f.data.copy_(torch.from_numpy(kernel_psf))

    def forward(self, x):
        return self.seq(x)



import numpy
import scipy

# from numpy.typing import ArrayLike
from scipy.ndimage import minimum_filter
from skimage.feature import peak_local_max

from aydin.it.classic_denoisers.gaussian import calibrate_denoise_gaussian
from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class PeriodicNoiseSuppressionTransform(ImageTransformBase):
    """Periodic Noise Suppression

    Some images have a form of periodic noise that can be seen as strong
    peaks in their power spectral density. Suppressing these peaks before and
    after denoising is often a good idea. This is tricky to use, use with
    care. Works with non-axis aligned periodic patterns.
    """

    preprocess_description = (
        "Periodic noise suppression" + ImageTransformBase.preprocess_description
    )
    postprocess_description = (
        "Reaply periodic noise" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = False

    def __init__(
        self,
        mask_radius: float = 0.01,
        threshold: float = 0.1,
        max_num_peaks: int = 32,
        post_processing_is_inverse: bool = False,
        priority: float = 0.30,
        **kwargs,
    ):

        """
        Constructs a Periodic Noise Suppression Transform

        Parameters
        ----------
        mask_radius : float
            Radius for masking peaks in Fourier spectrum. Expressed in units
            relative to the size of the image, i.e. must be within [0,1].
            Will always be at least a single pixel wide.
        threshold : float
            Threshold for FFT peak detection.
        max_num_peaks : int
            Maximum number of peaks in the Fourier spectra to correct for.
        post_processing_is_inverse : bool
            If False, we reapply the periodic suppression to eliminate any
            remaining periodic noise after denoising.
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)
        self.mask_radius = mask_radius
        self.threshold = threshold
        self.max_num_peaks = max_num_peaks
        self.post_processing_is_inverse = post_processing_is_inverse
        self._original_dtype = None
        self._coordinates = {}

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_original_dtype']
        del state['_coordinates']
        return state

    def __str__(self):
        return (
            f'{type(self).__name__}'
            f' (mask_radius={self.mask_radius},'
            f' threshold={self.threshold},'
            f' post_processing_is_inverse={self.post_processing_is_inverse})'
        )

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):

        with lsection(
            f"Applies periodic noise suppression to array of shape: {array.shape} and dtype: {array.dtype}"
        ):
            new_array = self._suppress_periodic_patterns(array)
            return new_array

    def postprocess(self, array):

        if not self.do_postprocess:
            return array

        with lsection(
            f"Reapplies periodic noise to array of shape: {array.shape} and dtype: {array.dtype}"
        ):
            # turns out it is better to suppress the periodic patterns before and after:
            if self.post_processing_is_inverse:
                new_array = self._reapply_periodic_patterns(array)
            else:
                new_array = self._suppress_periodic_patterns(array)
            return new_array

    def _suppress_periodic_patterns(self, array):
        self._original_dtype = array.dtype
        array = array.astype(numpy.float32, copy=False)

        # Compute Fourier intensity:
        ft = scipy.fft.fftn(array, workers=-1)
        ft = scipy.fft.fftshift(ft)
        spectrum = numpy.abs(ft).astype(numpy.float32)

        # Mask radius in pixels:
        mask_radius_pix = max(1, int(self.mask_radius * numpy.mean(array.shape)))

        # Denoise density:
        function, parameters, _ = calibrate_denoise_gaussian(
            spectrum, max_num_truncate=1
        )
        filtered_spectrum = function(spectrum, **parameters)

        # Minimum filtering:
        size = 2 * mask_radius_pix + 1
        filtered_spectrum = minimum_filter(filtered_spectrum, size=size)

        # Compute correction:
        correction = filtered_spectrum / spectrum

        # Coordinates of peaks to correct:
        self._coordinates = {}

        # Locate peaks:
        coordinates = peak_local_max(
            numpy.log1p(spectrum),
            min_distance=2 * mask_radius_pix,
            threshold_rel=self.threshold,
            num_peaks=self.max_num_peaks,
        )

        # center peak:
        center = tuple(s // 2 for s in array.shape)

        # For each peak:
        for i, coordinate in enumerate(coordinates):
            coordinate = tuple(coordinate)
            if coordinate != center:
                lprint(f"Suppressing peak at location: {coordinate}")

                # We compute a mask:
                mask = _sphere(
                    correction.shape, radius=mask_radius_pix, position=coordinate
                )

                # Apply the mask:
                factor = correction[mask]
                ft[mask] *= factor

                # keep coordinates:
                self._coordinates[i] = (mask, factor)

        # Transform back:
        ft = scipy.fft.ifftshift(ft)
        new_array = scipy.fft.ifftn(ft, workers=-1)
        new_array = numpy.real(new_array)

        # import napari
        # with napari.gui_qt():
        #     viewer = napari.Viewer()
        #     viewer.add_image(array, name='array')
        #     viewer.add_image(numpy.log1p(spectrum), name='spectrum')
        #     viewer.add_image(numpy.log1p(filtered_spectrum), name='filtered_spectrum')
        #     viewer.add_image(numpy.log1p(correction), name='correction')
        #     ft = scipy.fft.fftshift(ft)
        #     viewer.add_image(numpy.log1p(numpy.abs(ft).astype(numpy.float32)), name='spectrum_after')
        #     viewer.add_image(new_array, name='new_array')

        return new_array

    def _reapply_periodic_patterns(self, array):
        array = array.astype(numpy.float32, copy=False)
        ft = scipy.fft.fftn(array, workers=-1)
        ft = scipy.fft.fftshift(ft)
        for mask, factor in self._coordinates.values():
            ft[mask] /= factor
        ft = scipy.fft.ifftshift(ft)
        new_array = scipy.fft.ifftn(ft, workers=-1)
        new_array = numpy.real(new_array)
        # cast back to original dtype:
        new_array = new_array.astype(self._original_dtype, copy=False)
        return new_array


def _sphere(shape, radius, position):
    # From : https://stackoverflow.com/questions/46626267/how-to-generate-a-sphere-in-3d-numpy-array/46626448

    # assume shape and position are both a 3-tuple of int or float
    # the units are pixels / voxels (px for short)
    # radius is a int or float in px
    semisizes = (radius,) * len(shape)

    # generate the grid for the support points
    # centered at the position indicated by position
    grid = [slice(-x0, dim - x0) for x0, dim in zip(position, shape)]
    position = numpy.ogrid[grid]
    # calculate the distance of all points from `position` center
    # scaled by the radius
    arr = numpy.zeros(shape, dtype=float)
    for x_i, semisize in zip(position, semisizes):
        # this can be generalized for exponent != 2
        # in which case `(x_i / semisize)`
        # would become `np.abs(x_i / semisize)`
        arr += (x_i / semisize) ** 2

    # the inner part of the sphere will have distance below 1
    return arr <= 1.0



# flake8: noqa

import numpy
import numpy as np
import scipy
from aydin.features.fast.fast_features import FastFeatureGenerator
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo(image, name):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """

    Log.enable_output = True

    image = image[0:510, 0:510]
    image = normalise(image.astype(np.float32))

    noisy = add_noise(image)
    kernel = numpy.array([[0.25, 0.5, 0.25]])
    noisy = scipy.ndimage.convolve(noisy, kernel, mode='mirror')

    # brute-force feature-exclusion based extended blindspot:
    generator = FastFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_line_features=True,
        include_spatial_features=True,
    )
    regressor = LGBMRegressor(patience=20, max_num_estimators=2048)

    itnbs = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    itwbs = ImageTranslatorFGR(
        feature_generator=generator, regressor=regressor, blind_spots='auto'
    )

    itnbs.train(noisy)
    denoised_nbs = itnbs.translate(noisy)

    itwbs.train(noisy)
    denoised_wbs = itwbs.translate(noisy)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised_nbs = numpy.clip(denoised_nbs, 0, 1)
    denoised_wbs = numpy.clip(denoised_wbs, 0, 1)

    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised_nbs = psnr(image, denoised_nbs)
    ssim_denoised_nbs = ssim(image, denoised_nbs)
    psnr_denoised_wbs = psnr(image, denoised_wbs)
    ssim_denoised_wbs = ssim(image, denoised_wbs)

    print("noisy      :", psnr_noisy, ssim_noisy)
    print("denoised no extended blind-spot  :", psnr_denoised_nbs, ssim_denoised_nbs)
    print("denoised with extended blind-spot:", psnr_denoised_wbs, ssim_denoised_wbs)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(denoised_nbs, name='denoised_nbs')
        viewer.add_image(denoised_wbs, name='denoised_wbs')


camera_image = camera()
demo(camera_image, "camera")
# lizard_image = lizard()
# demo(lizard_image, "lizard")
# pollen_image = pollen()
# demo(pollen_image, "pollen")
# dots_image = dots()
# demo(dots_image, "dots")
# newyork_image = newyork()
# demo(newyork_image, "newyork")



import time
from os.path import join

import numpy
import pytest
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, add_noise
from aydin.io.folders import get_temp_folder
from aydin.it.base import ImageTranslatorBase
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor
from aydin.regression.lgbm import LGBMRegressor
from aydin.regression.linear import LinearRegressor
from aydin.regression.nn import NNRegressor
from aydin.regression.rf import RFRegressor
from aydin.regression.sv import SVRegressor


@pytest.mark.heavy
def test_saveload_Linear():
    regressor = LinearRegressor()
    saveload(StandardFeatureGenerator(), regressor, min_psnr=19, min_ssim=0.73)


@pytest.mark.heavy
def test_saveload_RF():
    regressor = RFRegressor()
    saveload(StandardFeatureGenerator(), regressor, min_ssim=0.75)


@pytest.mark.heavy
def test_saveload_SVR():
    regressor = SVRegressor()
    saveload(StandardFeatureGenerator(), regressor, min_psnr=22, min_ssim=0.71)


@pytest.mark.heavy
def test_saveload_NN():
    regressor = NNRegressor(max_epochs=12)
    saveload(StandardFeatureGenerator(), regressor, min_psnr=21, min_ssim=0.73)


def test_saveload_CB():
    regressor = CBRegressor(max_num_estimators=256, min_num_estimators=64)
    saveload(StandardFeatureGenerator(), regressor, min_ssim=0.78)


@pytest.mark.heavy
def test_saveload_LGBM():
    regressor = LGBMRegressor(max_num_estimators=256)
    saveload(StandardFeatureGenerator(), regressor, min_ssim=0.79)


def saveload(generator, regressor, min_psnr=22, min_ssim=0.75):
    image = normalise(camera().astype(numpy.float32))
    noisy = add_noise(image)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    it.train(noisy, noisy)

    temp_file = join(get_temp_folder(), "test_it_saveload" + str(time.time()))
    print(temp_file)

    it.save(temp_file)
    del it

    loaded_it = ImageTranslatorBase.load(temp_file)

    denoised = loaded_it.translate(noisy)

    denoised = denoised.clip(0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy

    # if the line below fails, then the parameters of the image the lgbm regressohave   been broken.
    # do not change the number below, but instead, fix the problem -- most likely a parameter.

    assert psnr_denoised > min_psnr and ssim_denoised > min_ssim



import numpy
from scipy.stats import entropy


def random_sample_patches(input_img, patch_size, num_patch, adoption_rate=0.5):
    """
    This function outputs a list of slices that crops a part of the input_img (i.e. patch).
    Only patches with higher entropy in their intensity histogram are selected.

    Parameters
    ----------
    input_img
        input image that will be sampled with patches
    patch_size
        patch_size
    num_patch
        number of patches to be output
    adoption_rate
        The % of patches selected from the original population of patches

    Returns
    -------
    patch_slice
    """
    if type(patch_size) == int:
        if len(input_img.shape) == 4:
            patch_size = (patch_size, patch_size)
        if len(input_img.shape) == 5:
            patch_size = (patch_size, patch_size, patch_size)

    img_dim = input_img.shape
    patchs_per_img = numpy.ceil(
        numpy.ceil(num_patch / img_dim[0]) / adoption_rate
    ).astype(int)
    coordinates = numpy.asarray(img_dim[1:-1]) - numpy.asarray(patch_size)
    patchs_per_img = min(patchs_per_img, numpy.prod(coordinates + 1))
    hist_ind_all = []
    for k in range(img_dim[0]):
        ind_past = []
        hist_batch = []
        while len(hist_batch) < patchs_per_img:
            # Randomly choose coordinates from an image.
            ind = numpy.hstack(
                [k]
                + [
                    numpy.random.choice(
                        1 if coordinates[i] == 0 else coordinates[i], 1, replace=True
                    ).astype(int)
                    for i in range(coordinates.size)
                ]
            )
            # Check if the new patch is too close to the existing patches.
            if ind_past:
                if abs(ind_past - ind).max() <= coordinates.min() // 20:
                    continue
            ind_past.append(ind)
            # Extract image patch from the input image.
            if len(patch_size) == 2:
                img0 = input_img[
                    ind[0],
                    ind[1] : ind[1] + patch_size[0],
                    ind[2] : ind[2] + patch_size[1],
                    0,
                ]
            elif len(patch_size) == 3:
                img0 = input_img[
                    ind[0],
                    ind[1] : ind[1] + patch_size[0],
                    ind[2] : ind[2] + patch_size[1],
                    ind[3] : ind[3] + patch_size[2],
                    0,
                ]
            else:
                raise ValueError('Only 2D or 3D patches are applicable.')
            # Calculate histogram and entropy
            hist, _ = numpy.histogram(img0, range=(0, 1), bins=255, density=True)
            hist_batch.append(entropy(hist))
        # Create a table with entropy and indices of each patch.
        hist_ind = numpy.hstack((numpy.vstack(hist_batch), ind_past))
        # Sort by entropy.
        hist_ind = hist_ind[(-hist_ind[:, 0]).argsort()]
        # Only leave the highest `adoption_rate` of patches.
        hist_ind = hist_ind[: max(int(hist_ind.shape[0] * adoption_rate), 1), ...]
        hist_ind_all.append(hist_ind)
    hist_ind_all = numpy.vstack(hist_ind_all)
    hist_ind_all = hist_ind_all[(-hist_ind_all[:, 0]).argsort()]
    hist_ind_all = hist_ind_all[: int(num_patch), 1:].astype(int)

    # Create a slice list
    patch_slice = []
    for ind in hist_ind_all:
        slice_list = (
            [slice(ind[0], ind[0] + 1, 1)]
            + [
                slice(ind[i + 1], ind[i + 1] + patch_size[i], 1)
                for i in range(len(patch_size))
            ]
            + [slice(0, 1, 1)]
        )
        patch_slice.append(tuple(slice_list))

    # Return a list of slice
    return patch_slice



# flake8: noqa
import time
from os.path import join

import napari
import numpy
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise
from aydin.io.folders import get_temp_folder
from aydin.it.base import ImageTranslatorBase
from aydin.it.cnn import ImageTranslatorCNN


def test_saveload_cnn():
    max_epochs = 2
    image_width = 200
    image = normalise(camera())
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    noisy = add_noise(image)

    it = ImageTranslatorCNN(
        # training_architecture='random',
        model_architecture='jinet',
        nb_unet_levels=2,
        mask_size=3,
        batch_norm='instance',
        max_epochs=max_epochs,
    )
    it.train(noisy, noisy)

    # values before loading
    denoised = it.translate(noisy, tile_size=image_width)
    denoised = denoised.reshape(image.shape)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    temp_file = join(get_temp_folder(), "test_it_saveload_cnn" + str(time.time()))
    print(f"savepath: {temp_file}")
    it.save(temp_file)
    del it

    loaded_it = ImageTranslatorBase.load(temp_file)

    denoised1 = numpy.clip(loaded_it.translate(noisy), 0, 1)
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(noisy)
    #     viewer.add_image(denoised)
    #     viewer.add_image(denoised1)
    psnr_denoised1 = psnr(denoised1.squeeze(), image)
    ssim_denoised1 = ssim(denoised1.squeeze(), image)
    print("denoised", psnr_denoised1, ssim_denoised1)

    assert 0.99 < psnr_denoised / psnr_denoised1 < 1.01
    assert 0.99 < ssim_denoised / ssim_denoised1 < 1.01



import napari
from qtpy.QtWidgets import QWidget, QPushButton, QApplication, QStyle, QHBoxLayout

from aydin.gui._qt.job_runners.worker import Worker
from aydin.it.fgr import ImageTranslatorFGR
from aydin.util.log.log import Log, lprint


class PreviewAllJobRunner(QWidget):
    def __init__(self, parent, threadpool):
        super(QWidget, self).__init__(parent)
        self.parent = parent
        self.threadpool = threadpool

        self.result_images = []
        self.preprocessed = []
        self.postprocessed = []

        self.widget_layout = QHBoxLayout()
        self.start_button = QPushButton(
            "Preview all", icon=QApplication.style().standardIcon(QStyle.SP_MediaPlay)
        )
        self.start_button.setFixedWidth(180)
        self.start_button.clicked.connect(self.prep_and_run)
        self.widget_layout.addWidget(self.start_button)

        self.setLayout(self.widget_layout)

    def start_func(self, progress_callback):
        Log.gui_callback = progress_callback

        for image in self.images:

            it = ImageTranslatorFGR()

            for transform in self.parent.transforms:
                transform_class = transform["class"]
                transform_kwargs = transform["kwargs"]
                it.add_transform(transform_class(**transform_kwargs))

            self.preprocessed.append(it._transform_preprocess_image(image))
            self.postprocessed.append(
                it._transform_postprocess_image(self.preprocessed[-1])
            )

        Log.gui_callback = None

    def progress_fn(self, log_str):
        self.parent.parent.activity_widget.infoTextBox.insertPlainText(log_str)

    def thread_complete(self):
        # Turn the overlay off
        # self.parent.overlay.hide()
        self.start_button.setEnabled(True)

        if self.preprocessed != [] and self.postprocessed != []:
            viewer = napari.Viewer()

            for image, preprocessed, postprocessed in zip(
                self.images, self.preprocessed, self.postprocessed
            ):

                viewer.add_image(image, name="image")
                viewer.add_image(preprocessed, name="preprocessed")
                viewer.add_image(postprocessed, name="postprocessed")

            viewer.grid.enabled = True
            viewer.grid.shape = (len(self.images), 3)
            viewer.show()

        self.preprocessed = []
        self.postprocessed = []

    def prep_and_run(self):
        # Get images and their related data
        self.images = self.parent.parent.tabs["Training Crop"].images
        if len(self.images) == 0:
            lprint("Preview All cannot be started with no image")
            return

        Log.gui_statusbar = self.parent.parent.parent.statusBar

        # Show activity widget
        self.parent.parent.activity_dock.setHidden(False)

        # Turn the overlay on
        # self.parent.parent.parent.parent.overlay.show()

        # Pass the function to execute
        worker = Worker(
            self.start_func
        )  # Any other args, kwargs are passed to the run function

        worker.signals.finished.connect(self.thread_complete)
        worker.signals.progress.connect(self.progress_fn)

        self.parent.parent.activity_widget.clear_activity()

        self.result_images = []

        self.start_button.setEnabled(False)

        # Execute
        self.threadpool.start(worker)



import numpy
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLabel, QTreeWidgetItem, QTreeWidget

from aydin.gui._qt.custom_widgets.horizontal_line_break_widget import (
    QHorizontalLineBreakWidget,
)
from aydin.gui._qt.qtreewidget_utils import iter_tree_widget
from aydin.util.misc.units import human_readable_byte_size


class ImagesTab(QWidget):
    """
    Next, we inspect the images contained in the added files.

    You can see basic image information such as name, axis label, array shapes, data types, volume in voxels,
    and size in bytes. When several images are loaded, each images is denoised independently which means that
    training and denoising are done per image. For bulk denoising of series of images please use the command line
    interface (CLI).
    """

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.parent = parent

        self.tab_layout = QVBoxLayout()

        # Explanation text
        self.explanation_text = QLabel(self.__doc__, self)
        self.tab_layout.addWidget(self.explanation_text)

        # Horizontal Line Break
        self.tab_layout.addWidget(QHorizontalLineBreakWidget(self))

        label = QLabel("Click on column header to check/uncheck all rows below.", self)
        self.tab_layout.addWidget(label)

        self.image_list_tree_widget = QTreeWidget()
        self.image_list_tree_widget.setHeaderLabels(
            ['file name', 'denoise', 'axes', 'shape', 'dtype', 'size']
        )

        self.image_list_tree_widget.header().sectionClicked.connect(
            self.onSectionClicked
        )
        self.image_list_tree_widget.header().setSectionsClickable(True)
        self.image_list_tree_widget.setColumnWidth(0, 400)
        self.image_list_tree_widget.setColumnWidth(3, 300)
        self.image_list_tree_widget.setColumnWidth(5, 200)
        self.image_list_tree_widget.itemChanged.connect(self.onTreeItemChanged)
        self.tab_layout.addWidget(self.image_list_tree_widget)

        self.tab_layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.tab_layout)

    @property
    def images2denoise(self):
        response = []
        root = self.image_list_tree_widget.invisibleRootItem()
        child_count = root.childCount()
        for i in range(child_count):
            item = root.child(i)
            response.append(bool(item.checkState(1)))

        return response

    @Slot(int)
    def onSectionClicked(self, column):
        if column == 0:
            return
        for idx, item in enumerate(
            iter_tree_widget(self.image_list_tree_widget.invisibleRootItem())
        ):
            if idx == 0:
                continue
            elif idx == 1:
                state_to_be_set = (
                    Qt.Unchecked
                    if item.checkState(column) == Qt.Checked
                    else Qt.Checked
                )

            item.setCheckState(column, state_to_be_set)

    def onTreeItemChanged(self, item, column):
        if column == 1:
            self.parent.data_model.set_image_to_denoise(
                item.text(0), item.checkState(1)
            )

    def remove_items_from_tree(self):
        self.image_list_tree_widget.clear()

    def setFollowingThreeTabsEnabled(self, bool):
        for _ in range(3, 6):
            self.parent.tabwidget.setTabEnabled(_, bool)

    def on_data_model_update(self):
        imagelist = self.parent.data_model.images
        if imagelist is None or imagelist == []:
            self.remove_items_from_tree()
            return

        self.image_list_tree_widget.clear()

        for filename, array, metadata, train_on, denoise, path in imagelist:
            qtree_widget_item = QTreeWidgetItem(
                self.image_list_tree_widget,
                [
                    filename,
                    ' ',
                    str(metadata.axes),
                    str(metadata.shape),
                    str(metadata.dtype),
                    human_readable_byte_size(
                        metadata.dtype.itemsize * numpy.prod(metadata.shape)
                    ),
                ],
            )
            qtree_widget_item.setCheckState(1, Qt.Checked if denoise else Qt.Unchecked)
            qtree_widget_item.setToolTip(0, path)



import numpy
import psutil
import tempfile
from typing import Tuple, Union, Generator

from aydin.util.log.log import lprint, lsection


class OffCore:
    memmap_directory = None


def offcore_array(
    shape: Union[Tuple[int, ...], Generator[int, None, None]],
    dtype: numpy.dtype,
    force_memmap: bool = False,
    zarr_allowed: bool = False,
    no_memmap_limit: bool = True,
    max_memory_usage_ratio: float = 0.9,
):
    """
    Instanciates an array of given shape and dtype in  'off-core' fashion i.e. not in main memory.
    Right now it simply uses memory mapping on temp file that is deleted after the file is closed

    Parameters
    ----------
    shape : tuple
    dtype : numpy.dtype
    force_memmap : bool
    zarr_allowed : bool
    no_memmap_limit : bool
    max_memory_usage_ratio : float

    """

    with lsection(f"Array of shape: {shape} and dtype: {dtype} requested"):
        size_in_bytes = numpy.prod(shape) * numpy.dtype(dtype).itemsize
        lprint(f'Array requested will be {(size_in_bytes / 1E6)} MB.')

        total_physical_memory_in_bytes = psutil.virtual_memory().total
        total_swap_memory_in_bytes = psutil.swap_memory().total

        total_mem_in_bytes = total_physical_memory_in_bytes + total_swap_memory_in_bytes
        lprint(
            f'There is {int(psutil.virtual_memory().total / 1E6)} MB of physical memory'
        )
        lprint(f'There is {int(psutil.swap_memory().total / 1E6)} MB of swap memory')
        lprint(f'There is {int(total_mem_in_bytes / 1E6)} MB of total memory')

        is_enough_total_memory = (
            size_in_bytes < max_memory_usage_ratio * total_mem_in_bytes
        )

        if not force_memmap and is_enough_total_memory:
            lprint(
                'There is enough physical+swap memory -- we do not need to use a mem mapped array or zarr-backed array.'
            )
            array = numpy.zeros(shape, dtype=dtype)

        elif no_memmap_limit:
            lprint(
                'There is not enough physical+swap memory -- we will use a mem mapped array.'
            )
            temp_file = tempfile.NamedTemporaryFile(dir=OffCore.memmap_directory)
            lprint(
                f'The temporary memory mapped file is at: {temp_file.name} (but you might not be able to see it!)'
            )
            array = numpy.memmap(temp_file, dtype=dtype, mode='w+', shape=shape)

        elif zarr_allowed:
            lprint(
                'There is not enough physical+swap memory -- we will use a zarr-backed array.'
            )
            import zarr

            array = zarr.create(
                shape=shape, dtype=dtype, store=zarr.TempStore("output.zarr")
            )
            # from numcodecs import Blosc
            # compressor = Blosc(cname = 'zstd', clevel = 3, shuffle = Blosc.BITSHUFFLE)
            # array = zarr.zeros((102_0, 200, 210), chunks = (100, 200, 210), compressor = compressor

        return array
